#define MAXBALL 100
#define MAXBLOCKED 1000

#define TRUE 1
#define FALSE 0

#include <stdio.h>

int size;
int robotposx, robotposy;
int ballposx[MAXBALL], ballposy[MAXBALL];
int goalposx[MAXBALL], goalposy[MAXBALL];
int ballcnt = 0;
int blockedx[MAXBLOCKED], blockedy[MAXBLOCKED];
int blockedcnt = 0;

void spec_err(void)
{
  fprintf(stderr,"Error in specification file! Exiting ...\n");
  exit(1);
}

void read_spec(FILE *f)
{
  printf("Read Spec ...\n");
  if (fscanf(f,"%d\n",&size) != 1) spec_err();
  printf("Size:    %d\n",size);
  if (fscanf(f,"%d %d\n",&robotposx,&robotposy) != 2) spec_err();
  printf("Robot:   x=%d y=%d\n",robotposx,robotposy);
  while (TRUE) {
    if (fscanf(f,"%d %d %d %d\n",&ballposx[ballcnt],&ballposy[ballcnt],
	       &goalposx[ballcnt],&goalposy[ballcnt]) != 4) spec_err();
    if (ballposx[ballcnt] != 0) {
      printf("From:    x=%d y=%d  To: x=%d y=%d\n",
	     ballposx[ballcnt],ballposy[ballcnt],
	     goalposx[ballcnt],goalposy[ballcnt]);
      ballcnt++;
      if (ballcnt >= MAXBALL) {
	fprintf(stderr,"Too many balls!\n");
	exit(1);
      }
    } else break;
  } 
  while (TRUE) {
    if (fscanf(f,"%d %d\n",&blockedx[blockedcnt],&blockedy[blockedcnt]) != 2) 
      spec_err();
    if (blockedx[blockedcnt] != 0) {
      printf("Blocked: x=%d y=%d\n",blockedx[blockedcnt],blockedy[blockedcnt]);
      blockedcnt++;
      if (blockedcnt >= MAXBLOCKED) {
	fprintf(stderr,"Too many blocked locations\n");
	exit(1);
      }
    } else break;
  }
}

void write_objs(FILE *f)
{
  int x,y,i,blocked;

  fprintf(f,"direction: up down left right;\n");
  fprintf(f,"ball: ");
  for (x=1; x <= ballcnt; x++) {
    if (x == ballcnt) fprintf(f,"  ball%1d;\n",x);
    else fprintf(f,"  ball%1d\n",x);
  }
  fprintf(f,"navigable: ");
  for (x=1; x <= size; x++) 
    for (y=1; y <= size; y++) {
      blocked = FALSE;
      for (i=0; i<blockedcnt; i++)
	if ((blockedx[i] == x) && (blockedy[i] == y)) blocked = TRUE;
      if (!blocked)
	fprintf(f,"a%1dx%1da ",x,y);
    }
  fprintf(f,";\n");
}



void write_adjs(FILE *f)
{
  int x,y;

  for (x=1; x <= size; x++)
    for (y=1; y <= size; y++) {
      if (x > 1) fprintf(f,"  adjacent(a%1dx%1da a%1dx%1da left)\n",
			 x,y,(x-1),y);
      if (x < size) fprintf(f,"  adjacent(a%1dx%1da a%1dx%1da right)\n",
			 x,y,x+1,y);
      if (y > 1) fprintf(f,"  adjacent(a%1dx%1da a%1dx%1da down)\n",
			 x,y,x,y-1);
      if (y < size) fprintf(f,"  adjacent(a%1dx%1da a%1dx%1da up)\n",
			    x,y,x,y+1);
    }
}

void write_initial(FILE *f)
{
  int i,x,y,empty;

  fprintf(f,"\n\ninitial: ");
  fprintf(f,"  at(robot a%1dx%1da)\n",robotposx,robotposy);
  for (i=0; i<ballcnt; i++)
    fprintf(f,"  at(ball%1d a%1dx%1da)\n",i+1,ballposx[i],ballposy[i]);
  for (x=1; x <= size; x++)
    for (y=1; y <= size; y++) {
      empty = TRUE;
      if ((robotposx == x) && (robotposy == y)) empty = FALSE;
      for (i=0; i <= ballcnt; i++)
	if ((ballposx[i] == x)  && (ballposy[i] == y)) empty = FALSE;
      if (empty) fprintf(f,"  empty(a%1dx%1da)\n",x,y);
    }
  write_adjs(f);
  fprintf(f,"  ;\n");
}

void write_goal(FILE *f)
{
  int i;

  fprintf(f,"\n\ngoal:");
  for (i=0; i < ballcnt; i++) 
    fprintf(f," at(ball%1d a%1dx%1da)",i+1,goalposx[i],goalposy[i]);
  fprintf(f,".\n");
}
      

void usage()
{
  fprintf(stderr,"sokoban <specfile> generates <specfile>_facts\n");
  exit(1);
}

void main(int argc, char *argv[])
{
  char factfile[1000];
  char specfile[1000];
  FILE *ff, *sf;

  if (argc != 2) usage();
  strcpy(specfile,argv[1]);
  strcpy(factfile,argv[1]);
  strcat(factfile,".fct");
  if ((sf = fopen(specfile,"r")) == NULL) {
    fprintf(stderr,"Cannot open spec file '%s' for reading\n");
    exit(1);
  }
  if ((ff = fopen(factfile,"w")) == NULL) {
    fprintf(stderr,"Cannot open fact file '%s' for writing\n");
    exit(1);
  }
  read_spec(sf);
  fclose(sf);

  write_objs(ff);
  write_initial(ff);
  write_goal(ff);
  fclose(ff);
}

  
  
